// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"hu","messages":{" add:":"hozz\u00e1ad:"," remove:":"elt\u00e1vol\u00edt:","$month$year":"$year$month","%(title)s (click to hide column)":"%(title)s (kattintson az oszlop elrejt\u00e9s\u00e9hez)","(empty)":"(\u00fcres)","(error)":"(hiba)","(no changeset information)":"(nincs v\u00e1ltoztat\u00e1s)","A filter already exists for that property":"A jellemz\u0151h\u00f6z m\u00e1r l\u00e9tezik sz\u0171r\u0151","Bold text: '''Example'''":"F\u00e9lk\u00f6v\u00e9r sz\u00f6veg: '''P\u00e9lda'''","Choose Time":"Id\u0151pont","Close":"Bez\u00e1r","Code block: {{{ example }}}":"K\u00f3dblokk: {{{ p\u00e9lda }}}","Done":"K\u00e9sz","Expand sub-directory in place":"Alk\u00f6nyvt\u00e1r kibont\u00e1sa helyben","Fold directory":"K\u00f6nyvt\u00e1r \u00f6sszehajt\u00e1sa","Heading: == Example ==":"Fejl\u00e9c: == P\u00e9lda ==","Horizontal rule: ----":"V\u00edzszintes elv\u00e1laszt\u00e1s: ----","Hour":"\u00d3ra","Image: [[Image()]]":"K\u00e9p: [[Image()]]","Italic text: ''Example''":"D\u0151lt sz\u00f6veg: ''P\u00e9lda''","Line break: [[BR]]":"Sort\u00f6r\u00e9s: [[BR]]","Link here":"Hivatkoz\u00e1s ide","Link to #%(id)s":"Hivatkoz\u00e1s erre: #%(id)s","Link to this diff":"Hivatkoz\u00e1s erre az elt\u00e9r\u00e9sre","Link to this section":"Hivatkoz\u00e1s erre a bekezd\u00e9sre","Link: [http://www.example.com/ Example]":"Hivatkoz\u00e1s: [http://www.example.com/ P\u00e9lda]","Loading %(entry)s...":"A(z) %(entry)s bet\u00f6lt\u00e9se...","Minute":"Perc","New paragraph":"\u00daj bekezd\u00e9s","Next":"K\u00f6vetkez\u0151","Now":"Most","Prev":"El\u0151z\u0151","Re-expand directory":"K\u00f6nyvt\u00e1r ism\u00e9telt kibont\u00e1sa","Second":"M\u00e1sodperc","Select ticket %(id)s for modification":"V\u00e1lassza a %(id)s jegyet a m\u00f3dos\u00edt\u00e1shoz","Show %(title)s":"A(z) %(title)s megjelen\u00edt\u00e9se","Tabular":"T\u00e1bl\u00e1zatos","Time":"Id\u0151","Time Zone":"Id\u0151z\u00f3na","Today":"Ma","Toggle selection of all tickets shown in this group":"Kijel\u00f6l\u00e9s megford\u00edt\u00e1sa az \u00f6sszes jegyen ebben a csoportban","Unified":"Egyben","Wk":"H\u00e9t","and":"\u00e9s","between":"k\u00f6z\u00f6tt","no":"nem","or":"vagy","yes":"igen"},"plural_expr":"(n!=1)"}).install();
