// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"da","messages":{" add:":"tilf\u00f8j:"," remove:":"fjern:","$month$year":"$month$year","%(title)s (click to hide column)":"%(title)s (klik for at skjule kolonne)","(empty)":"(tom)","(error)":"(fejl)","(no changeset information)":"(ingen rettes\u00e6t information)","A filter already exists for that property":"Der eksistere allerede et filter for den egenskab","Bold text: '''Example'''":"Fed tekst: ''Eksempel''","Choose Time":"V\u00e6lg tid","Close":"Luk","Code block: {{{ example }}}":"Kode blok: {{{ eksempel }}}","Done":"F\u00e6rdig","Expand sub-directory in place":"Udvid undermappe p\u00e5 stedet","Fold directory":"Sammenklap mappe","Heading: == Example ==":"Overskrift: == Eksempel ==","Horizontal rule: ----":"Horisontal linje: ----","Hour":"Time","Image: [[Image()]]":"Billede: [[Image()]]","Italic text: ''Example''":"Kursiv tekst: ''Eksempel''","Line break: [[BR]]":"Linjeskift: [[BR]]","Link here":"Link hertil","Link to #%(id)s":"Link til #%(id)s","Link to this diff":"Link til denne diff","Link to this section":"Link til denne sektion","Link: [http://www.example.com/ Example]":"Link: [http://www.example.com/ Eksempel]","Loading %(entry)s...":"Indl\u00e6ser %(entry)s...","Minute":"Minut","New paragraph":"Nyt afsnit","Next":"N\u00e6ste","Now":"Nu","Prev":"For","Re-expand directory":"Gen-udvid mappe","Second":"Sekund","Select ticket %(id)s for modification":"V\u00e6lg sag %(id)s der skal \u00e6ndres","Show %(title)s":"Vis %(title)s","Tabular":"Tabel","Time":"Tid","Time Zone":"Tidszone","Today":"I dag","Toggle selection of all tickets shown in this group":"Skift markering af alle sager der er vist i denne gruppe","Unified":"Samlet","Wk":"Uge","and":"og","between":"mellem","no":"nej","or":"eller","yes":"ja"},"plural_expr":"(n != 1)"}).install();
