<!--!  Copyright (C) 2006-2023 Edgewall Software

  This software is licensed as described in the file COPYING, which
  you should have received as part of this distribution. The terms
  are also available at https://trac.edgewall.org/wiki/TracLicense.

  This software consists of voluntary contributions made by many
  individuals. For the exact contribution history, see the revision
  history and logs, available at https://trac.edgewall.org/.
-->
<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:py="http://genshi.edgewall.org/"
      xmlns:i18n="http://genshi.edgewall.org/i18n"
      xmlns:xi="http://www.w3.org/2001/XInclude">
  <xi:include href="layout.html" />
  <head>
    <title>About Trac</title>
  </head>

  <body>
    <div id="content" class="about">

      <a id="trac-logo" class="logo" href="https://trac.edgewall.org/">
        <img src="${chrome.htdocs_location}trac_banner.png"
             alt="Trac: Integrated SCM &amp; Project Management" />
      </a>
      <h1>About Trac</h1>
      <p>Trac is a web-based software project management and bug/issue
        tracking system emphasizing ease of use and low ceremony.
        It provides an integrated Wiki, an interface to version control
        systems, and a number of convenient ways to stay on top of events
        and changes within a project.
      </p>
      <p i18n:msg="">Trac is distributed under the modified BSD License.<br />
        The complete text of the license can be found
        <a href="https://trac.edgewall.org/wiki/TracLicense">online</a>
        as well as in the <code>COPYING</code> file included in the distribution.</p>
      <a id="python-logo" class="logo" href="http://www.python.org/">
        <img src="${chrome.htdocs_location}python.png" alt="python powered" />
      </a>
      <p i18n:msg="">Please visit the Trac open source project:
        <a href="https://trac.edgewall.org/">https://trac.edgewall.org/</a></p>
      <p class="copyright" i18n:msg="year">Copyright &copy; ${'2003-2017'}
        <a href="http://www.edgewall.org/">Edgewall Software</a>
      </p>

      <xi:include href="environment_info.html" />

    </div>
  </body>
</html>
