// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"tr","messages":{"%(title)s (click to hide column)":"%(title)s (s\u00fctunu gizlemek i\u00e7in t\u0131klay\u0131n)","(empty)":"(bo\u015f)","(error)":"(hata)","(no changeset information)":"(de\u011fi\u015fimset bilgisi yok)","A filter already exists for that property":"Bu \u00f6zellik i\u00e7in zaten bir filtre mevcut","Bold text: '''Example'''":"Kal\u0131n metin: '''\u00d6rnek'''","Close":"Kapat","Code block: {{{ example }}}":"Kod blo\u011fu: {{{ \u00f6rnek }}}","Expand sub-directory in place":"Klas\u00f6r\u00fc geni\u015flet","Fold directory":"Klas\u00f6r\u00fc katla","Heading: == Example ==":"Ba\u015fl\u0131k: == \u00d6rnek ==","Horizontal rule: ----":"Yatay \u00e7izgi: ----","Image: [[Image()]]":"Resim: [[Image()]]","Italic text: ''Example''":"\u0130talik metin: ''\u00d6rnek''","Line break: [[BR]]":"Sat\u0131r sonu: [[BR]]","Link here":"Buraya ba\u011flant\u0131","Link to #%(id)s":"#%(id)s i\u00e7in ba\u011flant\u0131","Link to this diff":"Bu diff i\u00e7in ba\u011flant\u0131","Link to this section":"Bu b\u00f6l\u00fcme ba\u011flant\u0131","Link: [http://www.example.com/ Example]":"Ba\u011flant\u0131: [http://www.ornek.com/ \u00d6rnek]","Loading %(entry)s...":"Y\u00fckleniyor %(entry)s...","New paragraph":"Yeni paragraf","Re-expand directory":"Klas\u00f6r\u00fc yeniden geni\u015flet","Show %(title)s":"G\u00f6ster %(title)s","Tabular":"Tablolu","Unified":"Birle\u015fik","and":"ve","between":"aras\u0131nda","no":"hay\u0131r","or":"veya","yes":"evet"},"plural_expr":"0"}).install();
