// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"en_GB","messages":{" add:":" add:"," remove:":" remove:","$month$year":"$month$year","%(title)s (click to hide column)":"%(title)s (click to hide column)","(empty)":"(empty)","(error)":"(error)","(no changeset information)":"(no changeset information)","A filter already exists for that property":"A filter already exists for that property","Bold text: '''Example'''":"Bold text: '''Example'''","Choose Time":"Choose Time","Close":"Close","Code block: {{{ example }}}":"Code block: {{{ example }}}","Done":"Done","Expand sub-directory in place":"Expand sub-directory in place","Fold directory":"Fold directory","Heading: == Example ==":"Heading: == Example ==","Horizontal rule: ----":"Horizontal rule: ----","Hour":"Hour","Image: [[Image()]]":"Image: [[Image()]]","Italic text: ''Example''":"Italic text: ''Example''","Line break: [[BR]]":"Line break: [[BR]]","Link here":"Link here","Link to #%(id)s":"Link to #%(id)s","Link to this diff":"Link to this diff","Link to this section":"Link to this section","Link: [http://www.example.com/ Example]":"Link: [http://www.example.com/ Example]","Loading %(entry)s...":"Loading %(entry)s...","Minute":"Minute","New paragraph":"New paragraph","Next":"Next","Now":"Now","Prev":"Prev","Re-expand directory":"Re-expand directory","Second":"Second","Select ticket %(id)s for modification":"Select ticket %(id)s for modification","Show %(title)s":"Show %(title)s","Tabular":"Tabular","Time":"Time","Time Zone":"Time zone","Today":"Today","Toggle selection of all tickets shown in this group":"Toggle selection of all tickets shown in this group","Unified":"Unified","Wk":"Wk","and":"and","between":"between","no":"no","or":"or","yes":"yes"},"plural_expr":"(n != 1)"}).install();
