// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"ko","messages":{" add:":" \ucd94\uac00:"," remove:":" \uc0ad\uc81c:","$month$year":"$year\ub144 $month","%(title)s (click to hide column)":"%(title)s (\uc5f4\uc744 \uc228\uae30\ub824\uba74 \ud074\ub9ad)","(empty)":"(\ube44\uc5b4 \uc788\uc74c)","(error)":"(\uc624\ub958)","(no changeset information)":"(\ubcc0\uacbd\uc0ac\ud56d \uc815\ubcf4\uac00 \uc5c6\uc74c)","A filter already exists for that property":"\uc774 \uc18d\uc131\uc5d0 \ub300\ud55c \ud544\ud130\uac00 \uc774\ubbf8 \uc874\uc7ac\ud569\ub2c8\ub2e4.","Bold text: '''Example'''":"\uad75\uac8c: '''\uc608\uc81c'''","Choose Time":"\uc2dc\uac04 \uc120\ud0dd","Close":"\ub2eb\uae30","Code block: {{{ example }}}":"\ucf54\ub4dc \ube14\ub7ed: {{{ \uc608\uc81c }}}","Done":"\uc644\ub8cc","Expand sub-directory in place":"\uc774 \uc790\ub9ac\uc5d0 \ud558\uc704 \ub514\ub809\ud1a0\ub9ac \ud655\uc7a5","Fold directory":"\ub514\ub809\ud1a0\ub9ac \uc811\uae30","Heading: == Example ==":"\ud45c\uc81c: == \uc608\uc81c ==","Horizontal rule: ----":"\uac00\ub85c\uc904: ----","Hour":"\uc2dc","Image: [[Image()]]":"\uc774\ubbf8\uc9c0: [[Image()]]","Italic text: ''Example''":"\uae30\uc6b8\uc784\uaf34: ''\uc608\uc81c''","Line break: [[BR]]":"\uc904 \ubc14\uafc8: [[BR]]","Link here":"\uc5ec\uae30\ub85c \ub9c1\ud06c","Link to #%(id)s":"#%(id)s(\uc73c)\ub85c \ub9c1\ud06c","Link to this diff":"\uc774 \ucc28\uc774\uc810\uc73c\ub85c \ub9c1\ud06c","Link to this section":"\uc774 \ub2e8\ub77d\uc73c\ub85c \ub9c1\ud06c","Link: [http://www.example.com/ Example]":"\ub9c1\ud06c: [http://www.example.com/ \uc608\uc81c]","Loading %(entry)s...":"%(entry)s \ubd88\ub7ec\uc624\ub294 \uc911...","Minute":"\ubd84","New paragraph":"\uc0c8 \ub2e8\ub77d","Next":"\ub2e4\uc74c","Now":"\ud604\uc7ac","Prev":"\uc774\uc804","Re-expand directory":"\ub514\ub809\ud1a0\ub9ac \ub2e4\uc2dc \ud655\uc7a5","Second":"\ucd08","Select ticket %(id)s for modification":"\ubcc0\uacbd\uc744 \uc704\ud574 \ud2f0\ucf13 %(id)s\uc744(\ub97c) \uc120\ud0dd","Show %(title)s":"%(title)s \ud45c\uc2dc","Tabular":"\ud45c \ud615\uc2dd","Time":"\uc2dc\uac04","Time Zone":"\uc2dc\uac04\ub300","Today":"\uc624\ub298","Toggle selection of all tickets shown in this group":"\uc774 \uadf8\ub8f9\uc758 \ubaa8\ub4e0 \ud2f0\ucf13\uc5d0 \ub300\ud55c \uc120\ud0dd\uc744 \uc804\ud658","Unified":"\ud1b5\ud569","Wk":"\uc8fc","and":"\ub05d \uc2dc\uac04","between":"\uc2dc\uc791 \uc2dc\uac04","no":"\uc544\ub2c8\uc694","or":"\ub610\ub294","yes":"\uc608"},"plural_expr":"0"}).install();
