// Generated messages javascript file from compiled MO file
babel.Translations.load({"domain":"messages-js","locale":"es_MX","messages":{"%(title)s (click to hide column)":"%(title)s (clic para ocultar la columna)","(empty)":"(vac\u00edo)","(error)":"(error)","(no changeset information)":"(sin informaci\u00f3n de cambios)","A filter already exists for that property":"Ya existe un filtro para esa propiedad","Bold text: '''Example'''":"Texto en negrita: '''Ejemplo'''","Close":"Cerrar","Code block: {{{ example }}}":"Bloque de c\u00f3digo: {{{ ejemplo }}}","Expand sub-directory in place":"Desplegar subdirectorio aqu\u00ed","Fold directory":"Plegar directorio","Heading: == Example ==":"Encabezado: == Ejemplo ==","Horizontal rule: ----":"L\u00ednea horizontal: ----","Image: [[Image()]]":"Image:[[Image(tuimagen.ext)]]","Italic text: ''Example''":"Texto en cursiva: ''Ejemplo''","Line break: [[BR]]":"Forzar salto de l\u00ednea: [[BR]]","Link here":"Enlazar aqu\u00ed","Link to #%(id)s":"Enlazar a #%(id)s","Link to this diff":"Enlazar a esta diferencia","Link to this section":"Enlazar a esta secci\u00f3n","Link: [http://www.example.com/ Example]":"Enlace: [http://www.ejemplo.com/ Ejemplo]","Loading %(entry)s...":"Cargando %(entry)s...","New paragraph":"Nuevo p\u00e1rrafo","Re-expand directory":"Expandir directorio de nuevo","Show %(title)s":"Mostrar %(title)s","Tabular":"Tabla","Unified":"Unificado","and":"y","between":"entre","no":"no","or":"o","yes":"s\u00ed"},"plural_expr":"(n != 1)"}).install();
